
import os
import struct
import configparser

import matplotlib
matplotlib.use('AGG')#或者PDF, SVG或PS
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.ticker import MaxNLocator
from scipy.stats.stats import pearsonr
import torch.utils.data
from torch.utils.data import DataLoader
import pickle
import numpy as np
from sklearn.model_selection import train_test_split
import torch.nn.utils.rnn as rnn_utils
import re
import logging
import bisect


config = configparser.ConfigParser()
path_config=r'pDeepXL/configs/config.ini' # 在当前项目下，绝对路径要从项目根目录pDeepXL开始
# if os.path.isfile(path_config):
#     print('config file exist')
# else:
#     print('config file not exist')
config.read(path_config, encoding='UTF-8')
good_mods=set(config['DEFAULT']['valid_mods'].split(','))
# print('good mods: ', good_mods)

# ScanHeadsman-1.2.20200730输出的仪器名称和对应的简称
mpInstrumentShortNames={'Orbitrap Fusion Lumos':'Lumos','Q Exactive HF-X Orbitrap':'QEHFX','Orbitrap Fusion':'Fusion','Q Exactive HF Orbitrap':'QEHF','Q Exactive Orbitrap':'QE','Q Exactive Plus Orbitrap':'QEPlus','Exactive Plus Orbitrap':'QEPlus'}

# 读取仪器类型
def read_instrument_name(instrument_summary_path):
    fin=open(instrument_summary_path)
    lines=fin.readlines()
    fin.close()

    instrument_name_idx=8
    for i, col in enumerate(lines[0].split(',')):
        if col.strip()=='InstrumentName':
            instrument_name_idx = i

    # ans=lines[1].split(',')[8]
    contents=re.split(''',(?=(?:[^'"]|'[^']*'|"[^"]*")*$)''', lines[1].strip()) # 防止有些单元格内部逗号
    ans=contents[instrument_name_idx]
    return mpInstrumentShortNames[ans]

# instrument_summary_path=r'/data/zlchen/pDeepXL/data/PXD012546/R1_A5.sum.csv'
# print(read_instrument_name(instrument_summary_path))


# 判断一个字符串是否是数字，int和float都可以
def IsFloat(x):
    try:
        v=float(x)
        return True
    except ValueError:
        return False


# 读取每个MS2 Scan的能量
# 目前只支持单能量或三能量
def read_ms2_energy(energy_info_path):
    fin=open(energy_info_path)
    lines=fin.readlines()
    fin.close()

    hcd_energy_idx=0 # 不同质谱设置，能量所在的列不同，根据header行找到能量列
    energy_idx1=0 # 如果energy_idx对应列为负数，则用这一列
    activation_idx1=0 # 必须要是HCD碎裂模式
    for i, col in enumerate(lines[0].split(',')):
        if col.strip()=='HCD Energy': # 如果是阶梯能量，这一列显示最完整
            hcd_energy_idx = i
        elif col.strip()=='Energy1': # 如果是非阶梯能量，这一列也可以
            energy_idx1 = i
        elif col.strip()=='Activation1':
            activation_idx1 = i

    mpMS2Energy={}
    for i in range(1,len(lines)):
        # contents=lines[i].split(',')
        contents=re.split(''',(?=(?:[^'"]|'[^']*'|"[^"]*")*$)''', lines[i].strip()) # 有些单元格内部还有逗号
        msorder=int(contents[1])
        if msorder != 2:
            continue
        ms2scan=int(contents[0])

        energy_str=contents[energy_idx1]
        if contents[activation_idx1].strip().upper() != 'HCD': # 只提取HCD能量，单能量
            continue
        if '"' in contents[hcd_energy_idx]: # 多能量
            energy_str=contents[hcd_energy_idx]

        eg_contents=energy_str.split(',')
        if len(eg_contents) == 1: # 单能量
            mpMS2Energy[ms2scan]=[0.0, float(energy_str), 0.0] # low, medium, high
        elif len(eg_contents) == 3: # 三能量
            eg_contents=energy_str.strip()[1:-1].split(',') # 去掉外层的双引号
            mpMS2Energy[ms2scan]=sorted(list(map(float,eg_contents)))

    return mpMS2Energy


# energy_info_path=r'/data/zlchen/pDeepXL/data/PXD017620/CV2_DSSO_CV5_20200512_stepHCD_8_30B_R1.csv'
# energy_info_path=r'/data/zlchen/pDeepXL/data/PXD017620/XL/pf2/C_Lee_010916_ymitos_BS3_XL_A13_A15_11_Rep2.csv'
# energy_info_path=r'/data/zlchen/pDeepXL/data/PXD017620/crude_fresh/pf2/A_Linden_131118_211118_ymitos_crude_fresh_1to1_12.csv'
# energy_info_path=r'/data/zlchen/pDeepXL/data/PXD012546/R1_A5.csv'
# mpMS2Energy=read_ms2_energy(energy_info_path)
# print(mpMS2Energy)
# print('hello')


# 提取文件夹内所有文件的绝对路径
def absoluteFilePaths(directory):
    ans=[]
    for dirpath,_,filenames in os.walk(directory):
        for f in filenames:
            ans.append(os.path.abspath(os.path.join(dirpath, f)))
    return ans



# 读取pf2文件
def load_pf2(pf2_path):

    try:
        f = open(pf2_path, "rb")
        spec_title = os.path.basename(pf2_path)[:-10]
        nSpec,lenTitle = struct.unpack("2i", f.read(8))
        pf2title = struct.unpack("%ds" %lenTitle, f.read(lenTitle))
    except BaseException as error:
        print('An exception occurred: {}'.format(error))
        print('error pf2=%s'%pf2_path)
    #print(pf2title)
    
    mpSpec = {}
    
    for __i__ in range(nSpec):
        scan_no, = struct.unpack("i",f.read(4))
    #    print(">>Spectrum NO = %d" %scan_no)
        nPeak, = struct.unpack("i", f.read(4))
        peaks = []
        mz_int = struct.unpack(str(nPeak*2)+"d", f.read(nPeak*2*8))
        for i in range(nPeak):
            mz = mz_int[i*2]
            inten = mz_int[i*2 + 1]
            peaks.append( (mz, inten) )
        
        peaks=sorted(peaks,key=lambda x:x[0]) # 按mz从小到大排序，也许pf2默认已经排过序了。
        max_inten=0
        if len(peaks)!=0: # 当谱图没有任何谱峰时，会出bug：B170414_02_Lumos_ML_IN_205_FOMixSEC29BS3_TrypGluC_pepSECFr16.33307.33307.3.0.dta
            max_inten=max(peaks,key=lambda x:x[1])[1]
        
        nMix, = struct.unpack("i", f.read(4))
        nMaxCharge = 0
        for i in range(nMix):
            precursor, = struct.unpack("d", f.read(8))
            nCharge, = struct.unpack("i", f.read(4))
            if nCharge > nMaxCharge: nMaxCharge = nCharge
            specname = "%s.%d.%d.%d.%d.dta" %(spec_title, scan_no, scan_no, nCharge, i)
            spec_info=(specname, nCharge, precursor,max_inten)
            mpSpec[specname]=[spec_info,peaks]
    f.close()
    
    return pf2title, mpSpec


# 读取pFind3已按FDR过滤的结果
def load_pfind3_filtered(pfind3_path):
    
    fin=open(pfind3_path)
    lines=fin.readlines()
    fin.close()
    
    psms=[]

    for i in range(1,len(lines)):
        contents=lines[i].split('\t')
        title=contents[0].strip()
        scan=int(contents[1])
        exp_mh=float(contents[2])
        charge=int(contents[3])
        qvalue=float(contents[4])
        
        seq=contents[5].strip()
        ms=contents[10].strip().split(';')
        mods2={}
        for m in ms:
            if m=='':
                continue
            pos,name=m.split(',')
            pos=max(0, int(pos)-1) # 转换，修饰的下标从0开始；如果是N端修饰，则本身是0，不能出现负数
            name=name.strip()
            mods2[pos]=name

        psms.append([title,scan,charge,exp_mh,seq,mods2])
        
    return psms


# 读取pLink2过滤后的CSM结果
def ReadpLinkFilteredPSM5(strPLinkFilteredPath):
    vpLinkPSM = [] # title,scan,charge,exp_mh,seq1,mods1,linksite1,seq2,mods2,linksite2
    fpLink = open(strPLinkFilteredPath)
    lines=fpLink.readlines()
    fpLink.close()
    for i in range(1,len(lines)):
        line = lines[i].strip()
        contents = line.split(',')

        title = contents[1].strip()
        scan = int(title.split('.')[1])
        charge = int(contents[2])
        exp_mh = float(contents[3])

        pep1,pep2 = contents[4].split('-')
        seq1 = ''.join([i for i in pep1 if i.isalpha()])
        seq2 = ''.join([i for i in pep2 if i.isalpha()])
        lseq1=len(seq1)

        linksite1 = ''.join([i for i in pep1 if i.isnumeric()])
        linksite2 = ''.join([i for i in pep2 if i.isnumeric()])

        linksite1=max(0,int(linksite1)-1)
        linksite2=max(0,int(linksite2)-1)

        LinkerName = contents[6].strip()
        if LinkerName=='SS':
            assert (seq1[linksite1]=='C') and (seq2[linksite2]=='C') # 交联位点没问题
        else:
            assert (linksite1==0 or seq1[linksite1]=='K') and (linksite2==0 or seq2[linksite2]=='K') # 交联位点没问题

        mods=contents[8].strip().split(';')
        mods1,mods2={},{}

        for m in mods:
            if m.strip() == '' or m.lower() == 'null':
                continue
            mname,msite=m.split('(')
            msite=int(msite[:-1])
            maa=mname[-2:-1]

            if msite > lseq1+1: # beta修饰
                msite=max(0,msite-lseq1-3-1)
                assert seq2[msite]==maa
                mods2[msite]=mname
            else:
                msite=max(0,msite-1)
                assert seq1[msite] == maa
                mods1[msite]=mname

        vpLinkPSM.append([title,scan,charge,exp_mh,seq1,mods1,linksite1,seq2,mods2,linksite2])
    return vpLinkPSM



# 读取aa.ini
def readaaini(path):
    fin=open(path)
    lines=fin.readlines()
    fin.close()
    
    mpaa={}
    for line in lines:
        if line[0]=='R':
            mpcomp={}
            aa,com=line.split('=')[1].strip().split('|')[0:2]
            eles=com.split(')')
            for i in range(len(eles)-1):
                e,n=eles[i].split('(')
                n=int(n)
                mpcomp[e]=n
            mpaa[aa]=mpcomp
    return mpaa


# 读取element.ini
def readeleini(path):
    fin=open(path)
    lines=fin.readlines()
    fin.close()
    
    mpele={}
    for line in lines:
        if line[0]=='E':
            ele,mass,prob=line.split('=')[1].strip().split('|')[0:3]
            ms=[float(i) for i in mass.split(',')[:-1]]
            ps=[float(i) for i in prob.split(',')[:-1]]
            comb=list(zip(ms,ps))
            high_prob_mass=sorted(comb,key=lambda x:x[1])[-1][0]
            mpele[ele]=high_prob_mass
    return mpele



# 计算ppm质量误差
def calppm(theo,exp):
    return abs(1e6*(exp-theo)/theo) # 一定要加abs！！！！！！！！！！！

# 线性查找，太慢了！！！
def ismatch(spec,theo_mz,charge):
    if theo_mz<0:
        return 0.0
    spec_info,peaks = spec
    for exp_mz,intensity in peaks:
        if calppm(theo_mz*charge,exp_mz*charge)<=float(config['DEFAULT']['ppm_ms2']):
            return intensity
    return 0.0

# 二分查找，相比线性查找，提速至少30倍
# 经过单肽和交联数据集验证，结果和线性查找完全一样
def ismatch_bs(spec,theo_mz,charge):
    if theo_mz<0:
        return 0.0
    spec_info,peaks = spec
    n=len(peaks)
    exp_mzs=[p[0] for p in peaks] # 看load_pf2函数，已经是从小到大排序过的
    theo_mz_low,theo_mz_high=theo_mz-1,theo_mz+1 # 1Da的误差肯定比20ppm大
    low_idx,high_idx=bisect.bisect_left(exp_mzs,theo_mz_low),bisect.bisect_right(exp_mzs,theo_mz_high)
    for i in range(low_idx, high_idx+1):
        if i >=0 and i <n:
            exp_mz,intensity=peaks[i]
            if calppm(theo_mz*charge,exp_mz*charge)<=float(config['DEFAULT']['ppm_ms2']):
                return intensity
    return 0.0


def matchbatch(spec,ion_mzs,ion_charge):
    n=len(ion_mzs)
    mints=[0.0]*n
    for i,ion_mz in enumerate(ion_mzs):
        curint=ismatch_bs(spec,ion_mz,ion_charge)
        mints[i]=curint
    return mints


# 目前版本仅考虑C+57和O+16修饰
def is_mod_valid(mods):
    for msite,mname in mods.items():
        if mname not in good_mods:
            return False
    return True


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)

def epoch_time(start_time, end_time):
    elapsed_time = end_time - start_time
    elapsed_mins = int(elapsed_time / 60)
    elapsed_secs = int(elapsed_time - (elapsed_mins * 60))
    return elapsed_mins, elapsed_secs

def PlotWithEpoch(train_values, valid_values, path_result_home, name):
    epochs = [i + 1 for i in range(len(train_values))]

    plt.figure()
    plt.plot(epochs, train_values,'b',label='Training %s'%name)
    plt.plot(epochs, valid_values,'r',label='Validation %s'%name)
    plt.title('Training and Validation %s'%name)
    plt.legend()
    plt.xlim(1, len(epochs) + 1)

    ax = plt.gca()
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    plt.xlabel('epochs')
    plt.savefig(r'%s/%s.png'%(path_result_home,name))

def CountAccuPercentage(sims):
    s=sorted(sims)
    n=len(s)
    x=[s[0]] # 最小pcc
    y=[1.0] # yi=pcc>=xi的比例
    for i,pcc in enumerate(s):
        x.append(pcc)
        y.append(1-i/n)
    x.append(s[-1])
    y.append(1/len(sims)) # 最后一个
    return x,y

# input is sorted pccs and the corresponding perncentages
def BatchBSWantedPccs(pccs, percentages, wanted_pccs):
    wanted_percentages=[]
    for wpcc in wanted_pccs:
        idx=bisect.bisect_left(pccs, wpcc)
        if idx==len(pccs):
            wanted_percentages.append(0) # 所有pcc都没超过wpcc
        else:
            wanted_percentages.append(percentages[idx])
    return wanted_percentages


def PlotSimPercentage(train_sims, valid_sims, path_result_home, name):
    train_pcc,train_percentage=CountAccuPercentage(train_sims)
    valid_pcc,valid_percentage=CountAccuPercentage(valid_sims)

    plt.figure()
    plt.plot(train_pcc, train_percentage,'b',label='Training %s'%name)
    plt.plot(valid_pcc, valid_percentage,'r',label='Validation %s'%name)

    plt.title('Traing and Validation %s'%name)
    plt.legend()

    plt.xlabel('PCC = x')
    plt.ylabel('Percentage of PCCs >= x (%)')
    plt.savefig(r'%s/%s.png'%(path_result_home,name))


def PloSingleSimPercentage(pccs, percentages, path_result_home, name):
    plt.figure()
    plt.plot(pccs, percentages,'b',label=name)

    plt.title(name)
    plt.legend()

    plt.xlabel('PCC = x')
    plt.ylabel('Percentage of PCCs >= x (%)')
    plt.savefig(r'%s/%s.png'%(path_result_home,name))


# 输入numpy二维矩阵，shape：len*16
# 提取单肽离子，展开成一维向量
def flat_linear_pep_ions(input_matrix):
    return input_matrix[:,:4].flatten()

# # 输入numpy二维矩阵，shape：len*16
# # 提取不可碎裂交联肽段产生的离子
# # 包括单肽离子和xlink离子，展开成一维向量
# def flat_non_clv_pep_ions(input_matrix,l1,l2,s1,s2):
#     # 把交联输出拷贝到对应单肽位置
#     input_matrix[s1:l1,0:2]=input_matrix[s1:l1,4:6] # alpha cp xlink b
#     input_matrix[0:s1,2:4]=input_matrix[0:s1,6:8] # alpha cp xlink y
#     input_matrix[l1+1+s2:,0:2]=input_matrix[l1+1+s2:,4:6] # beta cp xlink b
#     input_matrix[l1+1:l1+1+s2,2:4]=input_matrix[l1+1:l1+1+s2,6:8] # beta cp xlink y
#     # 截取粘贴了交联离子的单肽区域出来算PCC
#     return flat_linear_pep_ions(input_matrix)

# 输入numpy二维矩阵，shape：len*16
# 提取可碎裂交联肽段产生的离子
# 包括单肽离子和clv-long、clv-short离子，展开成一维向量
def flat_clv_pep_ions(input_matrix,l1,l2,s1,s2):
    # 先把clv-long交联输出拷贝到对应单肽位置
    # input_matrix[s1:l1,0:2]=input_matrix[s1:l1,8:10] # alpha cp clv-long b
    # input_matrix[0:s1,2:4]=input_matrix[0:s1,10:12] # alpha cp clv-long y
    # input_matrix[l1+1+s2:,0:2]=input_matrix[l1+1+s2:,8:10] # beta cp clv-long b
    # input_matrix[l1+1:l1+1+s2,2:4]=input_matrix[l1+1:l1+1+s2,10:12] # beta cp clv-long y

    # 截取粘贴了交联离子的单肽区域出来，包括linear+clv_long
    linear_clv_long=flat_linear_pep_ions(input_matrix)

    # 然后提取clv-short匹配强度，把clv-short的y放到对应b的位置上，因为b位置无法匹配
    input_matrix[0:s1,4:6]=input_matrix[0:s1,6:8] # alpha cp clv-short y
    input_matrix[l1+1:l1+1+s2,4:6]=input_matrix[l1+1:l1+1+s2,6:8] # beta cp clv-short y
    # 把xlink的b区域提取出来
    clv_short=input_matrix[:,4:6].flatten()

    return np.append(linear_clv_long,clv_short)


# total_lens, len1s, len2s表示输入向量的总长度、pep1向量长度、pep2向量长度
# 分别=(l1-1)+(l2-1)+1、l1-1、l2-1，即比肽段实际长度短1
# 而linksite1s, linksite2s里面存储的是根据肽段实际长度确定的交联位点（下标从0开始）
def CalSim(predictions, truths, total_lens, pep_types, len1s, len2s, linksite1s, linksite2s):
    sims=[]
    for y_pred,y_truth,total_len,pep_type,l1,l2,s1,s2 in zip(predictions, truths, total_lens, pep_types, len1s, len2s, linksite1s, linksite2s):

        packed_y_pred=y_pred[:total_len].detach().cpu().numpy()
        packed_y_truth=y_truth[:total_len].detach().cpu().numpy()

        if pep_type==0 or pep_type==1: # linear pep or non-clv pep pair
            packed_y_pred=flat_linear_pep_ions(packed_y_pred)
            packed_y_truth=flat_linear_pep_ions(packed_y_truth)
        # elif pep_type==1: # non-clv pep pair
        #     packed_y_pred=flat_non_clv_pep_ions(packed_y_pred,l1,l2,s1,s2)
        #     packed_y_truth=flat_non_clv_pep_ions(packed_y_truth,l1,l2,s1,s2)
        else: # clv pep pair
            packed_y_pred=flat_clv_pep_ions(packed_y_pred,l1,l2,s1,s2)
            packed_y_truth=flat_clv_pep_ions(packed_y_truth,l1,l2,s1,s2)

        cursim=pearsonr(packed_y_pred,packed_y_truth)[0]
        sims.append(cursim)
    return sims


# https://vodkazy.cn/2019/12/12/%E5%B0%8F%E8%AE%B0%EF%BC%9A%E5%A4%84%E7%90%86LSTM-embedding%E5%8F%98%E9%95%BF%E5%BA%8F%E5%88%97/
class MyData(torch.utils.data.Dataset):
    def __init__(self, data_seq, data_label, data_title, data_type, data_l1, data_l2, data_s1, data_s2):
        self.data_seq = data_seq
        self.data_label = data_label
        self.data_title = data_title
        self.data_type = data_type
        self.data_l1 = data_l1
        self.data_l2 = data_l2
        self.data_s1 = data_s1
        self.data_s2 = data_s2

    def __len__(self):
        return len(self.data_seq)
    def __getitem__(self, idx):
        return self.data_seq[idx], self.data_label[idx], self.data_title[idx], self.data_type[idx], self.data_l1[idx], self.data_l2[idx], self.data_s1[idx], self.data_s2[idx]

def collate_fn(data):
    data.sort(key=lambda x: len(x[0]), reverse=True) # pack_padded_sequence 要求要按照序列的长度倒序排列
    x = [torch.Tensor(i[0]) for i in data]
    y = [torch.Tensor(i[1]) for i in data]
    x_padded = rnn_utils.pad_sequence(x, batch_first=True, padding_value=0)
    y_padded = rnn_utils.pad_sequence(y, batch_first=True, padding_value=0)

    xy_length,t_vec,pt_vec,l1_vec,l2_vec,s1_vec,s2_vec=[],[],[],[],[],[],[]
    for d in data:
        xy_length.append(len(d[0]))
        t_vec.append(d[2])
        pt_vec.append(d[3])
        l1_vec.append(d[4])
        l2_vec.append(d[5])
        s1_vec.append(d[6])
        s2_vec.append(d[7])
    return x_padded, y_padded, xy_length,t_vec,pt_vec,l1_vec,l2_vec,s1_vec,s2_vec


# def LoadDataSet(path_match_formatted_pkl, batch_size):
#     fformattedpkl = open(path_match_formatted_pkl, 'rb')
#     T,L,X,Y = pickle.load(fformattedpkl)
#     fformattedpkl.close()

#     X = np.array(X)
#     Y = np.array(Y)

#     logging.info('X.shape:%s'%str(X.shape))
#     logging.info('Y.shape:%s'%str(Y.shape))
    
#     X_train, X_test, Y_train, Y_test = train_test_split(X,Y, test_size=0.2, random_state=42)

#     train_data = MyData(X_train, Y_train)
#     train_data_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)

#     test_data = MyData(X_test, Y_test)
#     test_data_loader = DataLoader(test_data, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)

#     logging.info('full data size:%d'%len(X))
#     logging.info('train size:%d'%len(X_train))
#     logging.info('test size:%d'%len(X_test))

#     return train_data_loader, test_data_loader



def LoadDataSet(path_match_formatted_pkl, batch_size):
    fformattedpkl = open(path_match_formatted_pkl, 'rb')
    T,PT,L,L1,L2,S1,S2,X,Y = pickle.load(fformattedpkl)
    fformattedpkl.close()

    X = np.array(X)
    Y = np.array(Y)

    logging.info('X.shape:%s'%str(X.shape))
    logging.info('Y.shape:%s'%str(Y.shape))
    
    data = MyData(X, Y, T, PT, L1, L2, S1, S2)
    data_loader = DataLoader(data, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)

    logging.info('data size:%d'%len(X))

    return data_loader

################### for predict ################


# 读取单肽测试数据，数据格式和match_info.txt一致，只不过没有最后一列
def ReadpFind3PredictFile(pFind3_test_path):
    pass


# 读取CSM信息txt格式，包含匹配/预测强度信息
def ReadpLink2MatchInfoFile(pLink2_path):
    fin=open(pLink2_path)
    lines=fin.readlines()
    fin.close()

    int_idxs=[1,2,10,13]
    float_idxs=[4,5,6]
    dict_idxs=[9,12]
    list_idxs=[14,15]

    CSMs=[]
    mpTitleLines={}
    header=lines[0].strip()

    for i in range(1,len(lines)):
        contents=lines[i].strip().split('\t')
        mpTitleLines[contents[0]]=lines[i].strip() # 保留原始行信息

        for idx in int_idxs:
            contents[idx]=int(contents[idx])
        for idx in float_idxs:
            contents[idx]=float(contents[idx])
        for idx in dict_idxs:
            contents[idx]=eval(contents[idx])
        for idx in list_idxs:
            if idx < len(contents):
                contents[idx]=eval(contents[idx])

        CSMs.append(contents)
    return CSMs,mpTitleLines,header

# 读取交联测试数据,数据格式和match_info.txt一致，只不过没有最后两列
def ReadpLink2PredictFile(pLink2_path):
    CSMs,mpTitleLines,header=ReadpLink2MatchInfoFile(pLink2_path)

    new_CSMs=[]
    new_mpTitleLines={}
    new_header=header.split('\t')[:14]
    new_header='\t'.join(new_header)

    for csm in CSMs:
        new_CSMs.append(csm[:14])
    
    for title,line in mpTitleLines.items():
        new_mpTitleLines[title]='\t'.join(line.split('\t')[:14])
    
    return new_CSMs,new_mpTitleLines,new_header

# pLink2_path=r'/data/zlchen/pDeepXL/data/PXD019926/DSS/8PM/pLink2_data/pLink2_match_info_small_predict.txt'
# CSMs0,mpTitleLines0,header0=ReadpLink2PredictFile(pLink2_path)
# CSMs1,mpTitleLines1,header1=ReadpLink2PredictFile2(pLink2_path)
# print(CSMs0==CSMs1)
# print(mpTitleLines0==mpTitleLines1)
# print(header0==header1)
# pLink2_path=r'/data/zlchen/pDeepXL/data/PXD019926/DSS/8PM/pLink2_data/pLink2_match_info_small.txt'
# CSMs,mpTitleLines,header=ReadpLink2MatchInfoFile(pLink2_path)
# print(CSMs)
# print(mpTitleLines)
# print(header)





class MyPredictData(torch.utils.data.Dataset):
    def __init__(self, data_seq, data_title, data_type, data_l1, data_l2, data_s1, data_s2):
        self.data_seq = data_seq
        self.data_title = data_title
        self.data_type = data_type
        self.data_l1 = data_l1
        self.data_l2 = data_l2
        self.data_s1 = data_s1
        self.data_s2 = data_s2

    def __len__(self):
        return len(self.data_seq)
    def __getitem__(self, idx):
        return self.data_seq[idx], self.data_title[idx], self.data_type[idx], self.data_l1[idx], self.data_l2[idx], self.data_s1[idx], self.data_s2[idx]

def predict_collate_fn(data):
    data.sort(key=lambda x: len(x[0]), reverse=True) # pack_padded_sequence 要求要按照序列的长度倒序排列
    x = [torch.Tensor(i[0]) for i in data]
    x_padded = rnn_utils.pad_sequence(x, batch_first=True, padding_value=0)

    xy_length,t_vec,pt_vec,l1_vec,l2_vec,s1_vec,s2_vec=[],[],[],[],[],[],[]
    for d in data:
        xy_length.append(len(d[0]))
        t_vec.append(d[1])
        pt_vec.append(d[2])
        l1_vec.append(d[3])
        l2_vec.append(d[4])
        s1_vec.append(d[5])
        s2_vec.append(d[6])
    return x_padded, xy_length,t_vec,pt_vec,l1_vec,l2_vec,s1_vec,s2_vec


def ConvertPredictDataToDataLoader(batch_size, X, T, PT, L1, L2, S1, S2):
    predict_data = MyPredictData(X, T, PT, L1, L2, S1, S2)
    predict_data_loader = DataLoader(predict_data, batch_size=batch_size, shuffle=False, collate_fn=predict_collate_fn) # 测试数据不shuffle
    print('prediction size:', len(predict_data))
    return predict_data_loader





################## for fine tune #############
# 解析训练过程生成的日志文件

def ParseLineFromLog(line):
    ans=[]
    line=line.split('[INFO]')[1]
    contents=line.split(',')
    for part in contents:
        sep='=' if '=' in part else ':'
        k,v=part.split(sep)
        ans.append(float(v))
    return ans


def SmoothLog(log_info):
    min_idxs=[1,5]
    max_idxs=[3,4,6,7]
    smoothed=[]
    for i, contents in enumerate(log_info):
        if i==0:
            smoothed.append(contents)
        else:
            for idx in min_idxs:
                contents[idx]=min(contents[idx],smoothed[-1][idx])
            for idx in max_idxs:
                contents[idx]=max(contents[idx],smoothed[-1][idx])
            smoothed.append(contents)
    return smoothed

def ParseTrainLog(path_log,smooth=False):
    fin=open(path_log)
    lines=fin.readlines()
    fin.close()

    train_info=[]
    val_info=[]

    for line in lines:
        if 'Epoch:' in line:
            epoch=line.split('|')[0].split('Epoch:')[1].strip()
            epoch=int(epoch)
        elif 'Train Loss:' in line:
            cur_info=[epoch]+ParseLineFromLog(line)
            train_info.append(cur_info)
        elif 'Val. Loss:' in line:
            cur_info=[epoch]+ParseLineFromLog(line)
            val_info.append(cur_info)
    
    if smooth:
        return SmoothLog(train_info), SmoothLog(val_info)
    else:
        return train_info,val_info


# 批量画交联的肽谱匹配图

import os
import sys
sys.path.append("..")
import pDeepXL.utils as utils
import pDeepXL.mass_util as mass_util
import matplotlib
matplotlib.use('AGG')#或者PDF, SVG或PS
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.ticker import MaxNLocator
import numpy as np


# 不可碎裂
def plot_non_clv_single(title,pep_pair,prec_charge,pred_matrix,path_fig):
    LinkerName,seq1,mods1,linksite1,seq2,mods2,linksite2=pep_pair
    alpha_preds,beta_preds=pred_matrix
    alpha_preds,beta_preds=alpha_preds[:4],beta_preds[:4]
    
    l1,l2=len(seq1),len(seq2)

    b1mass1,b2mass1,y1mass1,y2mass1,b1mass2,b2mass2,y1mass2,y2mass2=mass_util.calfragmass4xl(seq1,mods1,linksite1,seq2,mods2,linksite2,LinkerName)

    alpha_ions=[b1mass1,b2mass1,y1mass1,y2mass1]
    beta_ions=[b1mass2,b2mass2,y1mass2,y2mass2]

    ion_charges=(1,2,1,2)
    ion_names=('b+','b++','y+','y++')
    ion_colors=('green','blue','red','orange')

    plt.figure(figsize=(15,8))

    min_mz,max_mz=10000,0

    # plot pred spec
    # alpha
    for ion_mzs, ion_charge, ion_name, ion_color,pred_intens in zip(alpha_ions, ion_charges, ion_names, ion_colors, alpha_preds):
        if prec_charge == 2 and ion_charge == 2:
            continue
        for pos,(theo_mz,pred_inten) in enumerate(zip(ion_mzs,pred_intens)):
            if pred_inten==0.0 or pred_inten==0 or theo_mz<=0.0:
                continue
            min_mz=min(min_mz,theo_mz)
            max_mz=max(max_mz,theo_mz)
            relative_intensity=pred_inten
            ion_pos= pos+1 if ion_name[0]=='b' else l1-pos
            ion_txt='α%s%d%s'%(ion_name[0],ion_pos,ion_name[1:])
            plt.plot([theo_mz,theo_mz], [0, relative_intensity], color=ion_color, lw=2)
            plt.text(theo_mz, relative_intensity + 0.03, ion_txt, rotation = 90, color=ion_color, horizontalalignment="center",verticalalignment='bottom')

    # beta
    for ion_mzs, ion_charge, ion_name, ion_color,pred_intens in zip(beta_ions, ion_charges, ion_names, ion_colors, beta_preds):
        if prec_charge == 2 and ion_charge == 2:
            continue
        for pos,(theo_mz,pred_inten) in enumerate(zip(ion_mzs,pred_intens)):
            if pred_inten==0.0 or pred_inten==0 or theo_mz<=0.0:
                continue
            min_mz=min(min_mz,theo_mz)
            max_mz=max(max_mz,theo_mz)
            relative_intensity=pred_inten
            ion_pos= pos+1 if ion_name[0]=='b' else l2-pos
            ion_txt='β%s%d%s'%(ion_name[0],ion_pos,ion_name[1:])
            plt.plot([theo_mz,theo_mz], [0, relative_intensity], color=ion_color, lw=2)
            plt.text(theo_mz, relative_intensity + 0.03, ion_txt, rotation = 90, color=ion_color, horizontalalignment="center",verticalalignment='bottom')

    mz_margin=20
    minx,maxx=min_mz-mz_margin,max_mz+mz_margin

    plt.plot([0,maxx],[0,0],color='black',lw=0.5)

    plt.title('predicted spectrum of %s'%title)
    # plt.legend()

    # plt.xlim(0,2800)
    plt.xlim(minx,maxx)
    plt.ylim(0, 1.2)

    # 修改纵坐标
    ax = plt.gca()
    labels = ['%d%%'%(i*10) for i in range(0,14,2)]
    ax.set_yticklabels(labels)

    plt.xlabel('m/z')
    plt.ylabel('Relative Abundance')
    # plt.savefig(path_fig+'.svg', format='svg', dpi=1200)
    plt.savefig(path_fig)


# 可碎裂
def plot_clv_single(title,pep_pair,prec_charge,pred_matrix,path_fig):
    LinkerName,seq1,mods1,linksite1,seq2,mods2,linksite2=pep_pair
    alpha_preds,beta_preds=pred_matrix
    
    l1,l2=len(seq1),len(seq2)

    # 把8维拆开为12维，和下面的ion_names对应
    alpha_preds=np.array(alpha_preds)
    alpha_preds_expended=np.zeros([12,l1])
    alpha_preds_expended[0:2,0:linksite1]=alpha_preds[0:2,0:linksite1] # regular b
    alpha_preds_expended[2:4,linksite1+1:]=alpha_preds[2:4,linksite1+1:] # regular y
    alpha_preds_expended[4:6,linksite1:]=alpha_preds[0:2,linksite1:] # clv-long b
    alpha_preds_expended[6:8,0:linksite1+1]=alpha_preds[2:4,0:linksite1+1] # clv-long y
    alpha_preds_expended[8:12,:]=alpha_preds[4:8,:] # clv-short
    alpha_preds=alpha_preds_expended.tolist()

    beta_preds=np.array(beta_preds)
    beta_preds_expended=np.zeros([12,l2])
    beta_preds_expended[0:2,0:linksite2]=beta_preds[0:2,0:linksite2] # regular b
    beta_preds_expended[2:4,linksite2+1:]=beta_preds[2:4,linksite2+1:] # regular y
    beta_preds_expended[4:6,linksite2:]=beta_preds[0:2,linksite2:] # clv-long b
    beta_preds_expended[6:8,0:linksite2+1]=beta_preds[2:4,0:linksite2+1] # clv-long y
    beta_preds_expended[8:12,:]=beta_preds[4:8,:] # clv-short
    beta_preds=beta_preds_expended.tolist()
    #################################


    alpha_ions=mass_util.calfragmass4clv(seq1,LinkerName,linksite1,mods1)
    beta_ions=mass_util.calfragmass4clv(seq2,LinkerName,linksite2,mods2)

    ion_charges=(1,2,1,2,1,2,1,2,1,2,1,2)
    #ion_names=('linear_b1','linear_b2','linear_y1','linear_y2','clv_long_b1','clv_long_b2','clv_long_y1','clv_long_y2','clv_short_b1','clv_short_b2','clv_short_y1','clv_short_y2')
    ion_names=('b+','b++','y+','y++','Lb+','Lb++','Ly+','Ly++','Sb+','Sb++','Sy+','Sy++')
    ion_colors=('green','blue','red','orange','green','blue','red','orange','green','blue','red','orange')

    plt.figure(figsize=(15,8))

    min_mz,max_mz=10000,0

    # plot pred spec
    # alpha
    for ion_mzs, ion_charge, ion_name, ion_color,pred_intens in zip(alpha_ions, ion_charges, ion_names, ion_colors, alpha_preds):
        if prec_charge == 2 and ion_charge == 2:
            continue
        for pos,(theo_mz,pred_inten) in enumerate(zip(ion_mzs,pred_intens)):
            if pred_inten==0.0 or pred_inten==0 or theo_mz<=0.0:
                continue
            min_mz=min(min_mz,theo_mz)
            max_mz=max(max_mz,theo_mz)
            relative_intensity=pred_inten
            ion_pos= pos+1 if 'b' in ion_name else l1-pos
            ion_txt='α%s%d%s'%(ion_name.split('+')[0],ion_pos,'+'*ion_charge)
            # print('type=%s,theomz=%f,intensity=%f'%(ion_txt,theo_mz,relative_intensity))
            plt.plot([theo_mz,theo_mz], [0, relative_intensity], color=ion_color, lw=2)
            plt.text(theo_mz, relative_intensity + 0.03, ion_txt, rotation = 90, color=ion_color, horizontalalignment="center",verticalalignment='bottom')

    # beta
    for ion_mzs, ion_charge, ion_name, ion_color,pred_intens in zip(beta_ions, ion_charges, ion_names, ion_colors, beta_preds):
        if prec_charge == 2 and ion_charge == 2:
            continue
        for pos,(theo_mz,pred_inten) in enumerate(zip(ion_mzs,pred_intens)):
            if pred_inten==0.0 or pred_inten==0 or theo_mz<=0.0:
                continue
            min_mz=min(min_mz,theo_mz)
            max_mz=max(max_mz,theo_mz)
            relative_intensity=pred_inten
            ion_pos= pos+1 if 'b' in ion_name else l2-pos
            ion_txt='β%s%d%s'%(ion_name.split('+')[0],ion_pos,'+'*ion_charge)
            # print('type=%s,theomz=%f,intensity=%f'%(ion_txt,theo_mz,relative_intensity))
            plt.plot([theo_mz,theo_mz], [0, relative_intensity], color=ion_color, lw=2)
            plt.text(theo_mz, relative_intensity + 0.03, ion_txt, rotation = 90, color=ion_color, horizontalalignment="center",verticalalignment='bottom')

    mz_margin=20
    minx,maxx=min_mz-mz_margin,max_mz+mz_margin
    plt.plot([0,maxx],[0,0],color='black',lw=0.5)

    plt.title('predicted spectrum of %s'%title)
    # plt.legend()

    # plt.xlim(50,1200)
    plt.xlim(minx,maxx)
    plt.ylim(0, 1.2)

    # 修改纵坐标
    ax = plt.gca()
    labels = ['%d%%'%(i*10) for i in range(0,14,2)]
    ax.set_yticklabels(labels)

    plt.xlabel('m/z')
    plt.ylabel('Relative Abundance')
    # plt.savefig(path_fig+'.svg', format='svg', dpi=1200)
    plt.savefig(path_fig)



def plot_batch(path_pLink2_match_info, path_img_folder):
    if not os.path.exists(path_img_folder):
        os.mkdir(path_img_folder)
    print('start plotting...')
    CSMs,mpTitleLines,header=utils.ReadpLink2MatchInfoFile(path_pLink2_match_info)
    for i,csm in enumerate(CSMs):
        title,scan,charge,instrument,NCE_low,NCE_medium,NCE_high,crosslinker,seq1,mods1,linksite1,seq2,mods2,linksite2,seq1_pred_b1b2y1y2,seq2_pred_b1b2y1y2=csm
        print('%d/%d plotting %s...'%(i,len(CSMs),title))
        pep_pair=[crosslinker,seq1,mods1,linksite1,seq2,mods2,linksite2]
        pred_matrix=seq1_pred_b1b2y1y2,seq2_pred_b1b2y1y2
        path_fig=r'%s/%s.png'%(path_img_folder,title)
        if crosslinker=='DSSO' or crosslinker=='DSBU':
            plot_clv_single(title,pep_pair,charge,pred_matrix,path_fig)
        elif crosslinker=='DSS' or crosslinker=='Leiker':
            plot_non_clv_single(title,pep_pair,charge,pred_matrix,path_fig)
        else:
            print('do not support %s cross-linker'%crosslinker)
    print('plot done.')



def plot_single(title,prec_charge,crosslinker,seq1,mods1,linksite1,seq2,mods2,linksite2,pred_matrix,path_fig):
    pep_pair=crosslinker,seq1,mods1,linksite1,seq2,mods2,linksite2
    if crosslinker=='DSSO' or crosslinker=='DSBU':
        plot_clv_single(title,pep_pair,prec_charge,pred_matrix,path_fig)
    elif crosslinker=='DSS' or crosslinker=='Leiker':
        plot_non_clv_single(title,pep_pair,prec_charge,pred_matrix,path_fig)
    else:
        print('do not support %s cross-linker'%crosslinker)




class PeptideFeaturer:
    # 系统参数
    MIN_PREC_CHARGE, MAX_PREC_CHARGE = 0, 0
    MIN_PEPTIDE_LEN, MAX_PEPTIDE_LEN = 0, 0
    MAX_NCE = 0
    VALID_MODS, VALID_INSTS, VALID_CROSS_LINKERS = [], [], []
    VALID_AA = 'ACDEFGHIKLMNPQRSTVWY-' # 最后一个字符-表示交联剂，用来分割alpha和beta序列

    def __init__(self, config):
        self.MIN_PREC_CHARGE=int(config['DEFAULT']['min_prec_charge'])
        self.MAX_PREC_CHARGE=int(config['DEFAULT']['max_prec_charge'])

        self.MIN_PEPTIDE_LEN=int(config['DEFAULT']['min_peptide_len'])
        self.MAX_PEPTIDE_LEN=int(config['DEFAULT']['max_peptide_len'])

        self.MAX_NCE=float(config['DEFAULT']['max_nce'])

        self.VALID_MODS=config['DEFAULT']['valid_mods'].split(',')
        self.VALID_INSTS=config['DEFAULT']['instruments'].split(',')
        self.VALID_CROSS_LINKERS=config['DEFAULT']['cross_linkers'].split(',')

    def IsSeqValid(self, seq):
        for aa in seq:
            if aa not in self.VALID_AA:
                return False
        return True

    def AA2OneHot(self, aa):
        ans=[0]*len(self.VALID_AA)
        ans[self.VALID_AA.index(aa)]=1
        return ans

    def Mod2OneHot(self, mname=''):
        ans=[0]*len(self.VALID_MODS)
        if mname =='':
            return ans
        ans[self.VALID_MODS.index(mname)]=1
        return ans

    def Charge2OneHot(self, c=0):
        ans=[0]*self.MAX_PREC_CHARGE
        if c==0:
            return ans
        ans[c-1]=1
        return ans

    def Instrument2OneHot(self, inst=''):
        ans=[0]*len(self.VALID_INSTS)
        if inst=='':
            return ans
        ans[self.VALID_INSTS.index(inst)]=1
        return ans

    def CrossLinker2OneHot(self, cross_linker=''):
        ans=[0]*len(self.VALID_CROSS_LINKERS)
        if cross_linker=='':
            return ans
        ans[self.VALID_CROSS_LINKERS.index(cross_linker)]=1
        return ans

    def PaddingZero(self, input_vec):
        n=len(input_vec)
        if n<self.MAX_PEPTIDE_LEN-1:
            for i in range(self.MAX_PEPTIDE_LEN-1-n):
                input_vec.append([0]*len(input_vec[0]))
        return input_vec

    def Add2ListElementWise(self,list1,list2):
        return [sum(x) for x in zip(list1, list2)]

    def LinkerSymbol2Vec(self):
        cur_aa_vec = self.AA2OneHot('-')
        cur_mod_vec = self.Mod2OneHot()
        left_vec = right_vec = left_sum_vec = right_sum_vec = cur_aa_vec + cur_mod_vec
        is_left_n_term, is_right_c_term = 0, 0
        is_left_link_site,is_right_link_site = 0, 0

        charge_vec = self.Charge2OneHot()
        inst_vec = self.Instrument2OneHot()
        cross_linker_vec = self.CrossLinker2OneHot()
        NCE_low,NCE_medium,NCE_high=0,0,0

        clv_vec = left_vec + right_vec + left_sum_vec + right_sum_vec + [is_left_n_term, is_right_c_term] + \
            [is_left_link_site, is_right_link_site] + charge_vec + inst_vec + [NCE_low,NCE_medium,NCE_high] + cross_linker_vec
            
        return clv_vec

    # 如果cross_linker=='',link_site==-1表示非交联
    def Sequence2Vec(self, seq, mod, charge, inst, NCE_low,NCE_medium,NCE_high, cross_linker='', link_site=-1, padding=False):
        NCE_low,NCE_medium,NCE_high=NCE_low/self.MAX_NCE,NCE_medium/self.MAX_NCE,NCE_high/self.MAX_NCE
        lseq = len(seq)
        aa_mod_vec = [] # 每个aa位置的aa+mod的vec
        for i in range(lseq):
            cur_aa_vec = self.AA2OneHot(seq[i])
            cur_mod_vec = self.Mod2OneHot()
            if i in mod:
                cur_mod_vec = self.Mod2OneHot(mod[i])

            cur_vec = cur_aa_vec + cur_mod_vec
            aa_mod_vec.append(cur_vec)

        # 左边aa的累加和，右边aa的累加和
        left_sum_aa_mod_vec, right_sum_aa_mod_vec = [0]*lseq, [0]*lseq
        for i in range(lseq):
            left_idx, right_idx = i, lseq-i-1
            cur_left_vec = aa_mod_vec[left_idx]
            cur_right_vec = aa_mod_vec[right_idx]

            if i == 0:
                left_sum_aa_mod_vec[left_idx]=cur_left_vec
                right_sum_aa_mod_vec[right_idx]=cur_right_vec
            else:
                left_sum_aa_mod_vec[left_idx]=self.Add2ListElementWise(cur_left_vec, left_sum_aa_mod_vec[left_idx-1])
                right_sum_aa_mod_vec[right_idx]=self.Add2ListElementWise(cur_right_vec, right_sum_aa_mod_vec[right_idx+1])

        # 最终向量
        seq_vec=[]
        for i in range(0, lseq-1):
            left_vec,right_vec=aa_mod_vec[i],aa_mod_vec[i+1]
            left_sum_vec,right_sum_vec=left_sum_aa_mod_vec[i],right_sum_aa_mod_vec[i+1]

            is_left_n_term, is_right_c_term = 0, 0
            if i == 0:
                is_left_n_term = 1
            elif i == lseq-2:
                is_right_c_term = 1

            is_left_link_site,is_right_link_site = 0, 0
            if i == link_site:
                is_left_link_site = 1
            elif i+1 == link_site:
                is_right_link_site = 1
            
            charge_vec = self.Charge2OneHot(charge)
            inst_vec = self.Instrument2OneHot(inst)
            cross_linker_vec = self.CrossLinker2OneHot(cross_linker)

            clv_vec = left_vec + right_vec + left_sum_vec + right_sum_vec + [is_left_n_term, is_right_c_term] + \
                [is_left_link_site, is_right_link_site] + charge_vec + inst_vec + [NCE_low,NCE_medium,NCE_high] + cross_linker_vec
            seq_vec.append(clv_vec)

        if padding:
            seq_vec = self.PaddingZero(seq_vec)

        return seq_vec, left_sum_aa_mod_vec[lseq-1] # 同时返回整个肽段的aa_mod累加和，检验过与right_sum_aa_mod_vec[0]相同


    def AddAnotherSeq(self, target_vec, link_site, another_seq_aa_mod_sum_vec):
        vec_len = len(target_vec)
        left_sum_st, left_sum_end = 52,78
        right_sum_st, right_sum_end = 78,104
        for i in range(link_site, vec_len):
            target_vec[i][left_sum_st:left_sum_end] = self.Add2ListElementWise(target_vec[i][left_sum_st:left_sum_end], another_seq_aa_mod_sum_vec)
        for i in range(0,link_site):
            target_vec[i][right_sum_st:right_sum_end] = self.Add2ListElementWise(target_vec[i][right_sum_st:right_sum_end], another_seq_aa_mod_sum_vec)
        return target_vec


if __name__ == "__main__":
    import configparser
    config = configparser.ConfigParser()
    config.read('/data/zlchen/pDeepXL/code/pDeepXL/config.ini')
    pf = PeptideFeaturer(config)
    # seq_vec = pf.Sequence2Vec('MIAKSEQEIGK',{0:'Oxidation[M]',4:'Carbamidomethyl[C]',5:'Carbamidomethyl[C]',7:'Oxidation[M]'},5,'Lumos',0,30,0,'',-1,False)
    seq_vec, left_sum, right_sum = pf.Sequence2Vec('MIAKSEQEIGK',{0:'Oxidation[M]',5:'Carbamidomethyl[C]',7:'Oxidation[M]'},3,'Lumos',0,30,0,'DSS',3,False)
    for clv_vec in seq_vec:
        print(clv_vec)
    print(left_sum==right_sum)
    print(left_sum)



import pDeepXL.utils as utils

paa=r'pDeepXL/configs/aa.ini'
pele=r'pDeepXL/configs/element.ini'

mpLinkerXLMass={'DSS':138.068,'Leiker':316.142,'SS':-2.016,'DSSO':158.004,'DSBU':196.085}
mpClvLinkerLongShortMass={'DSSO':[85.983,54.011],'DSBU':[111.032,85.053]}

mpModMassN14={'Carbamidomethyl[C]':57.021464,'Oxidation[M]':15.994915,'Deamidated[N]':0.984016,
           'Acetyl[ProteinN-term]':42.010565,'Sulfo-NHS-LC-LC-Biotin[AnyN-term]':452.245726,
           'LG-anhydrolactam[AnyN-term]':314.188195,'Biotin_Thermo-21330[AnyN-term]':473.219571,
           'NHS-LC-Biotin[AnyN-term]':339.161662,'LG-anhyropyrrole[AnyN-term]':298.193280,
           'Cytopiloyne+water[AnyN-term]':380.147118,'LG-Hlactam-K[ProteinN-term]':348.193674,
           'bisANS-sulfonates[T]':437.201774,'CLIP_TRAQ_3[AnyN-term]':271.148736,
           'Gln->pyro-Glu[AnyN-termQ]':-17.026547,'Nethylmaleimide[C]':125.047679} # 添加的修饰名称及质量

mpaa=utils.readaaini(paa)
mpele=utils.readeleini(pele)

n14aamass={}

for aa, formula in mpaa.items():
    n14mass=0.0
    for e,n in formula.items():
        n14mass+=(n*mpele[e])
    n14aamass[aa]=n14mass

H2OMass=mpele['H']*2+mpele['O']
PMass=1.0072766

# 计算含水的单肽完整质量
def calpepmass(seq,mods={}):
#    H2OMass=0.0
    n14mass=H2OMass
    for aa in seq:
        n14mass+=n14aamass[aa]
    
    for msite,mname in mods.items():
        n14mass+=mpModMassN14[mname]
    
    return n14mass


# seq='EKYIDQEELNK'
# mods=[]
# n14mass=calpepmass(seq,mods)
# print(n14mass)

# 计算含水的交联完整质量
def calxlpepmass(seq,LinkerName,mods={}):
#    H2OMass=0.0
    linear=''.join([i if i.isalpha() else '' for i in seq])
    n14mass=calpepmass(linear,mods)
    LinkerXLMass=mpLinkerXLMass[LinkerName]
    return n14mass+H2OMass+LinkerXLMass


# seq='DSGKELHINLIPNKQ(3)-EKYIDQEELNK(2)'
# mods=[]
# n14mass=calxlpepmass(seq,mods)
# print(n14mass)


# 计算单肽的碎片离子质量
def calfragmass4regular(seq,mods={}):
    n=len(seq)
    mass=[0.0]*n
    for i,aa in enumerate(seq):
        mass[i]=n14aamass[aa]
    for msite,mname in mods.items():
        mass[msite]+=(mpModMassN14[mname])
        
    b1mass=[0.0]*n
    b1mass[0]=mass[0]+PMass
    for i in range(1,n):
        b1mass[i]=b1mass[i-1]+mass[i]
    y1mass=[0.0]*n
    y1mass[n-1]=mass[n-1]+H2OMass+PMass
    for i in range(n-2,-1,-1):
        y1mass[i]=y1mass[i+1]+mass[i]
    

    b2mass=[0.0]*n
    b3mass=[0.0]*n
    y2mass=[0.0]*n
    y3mass=[0.0]*n
    for i in range(n):
        b2mass[i]=(b1mass[i]+PMass)/2
        b3mass[i]=(b1mass[i]+2*PMass)/3

        y2mass[i]=(y1mass[i]+PMass)/2
        y3mass[i]=(y1mass[i]+2*PMass)/3
        
    b1mass[n-1]=b2mass[n-1]=b3mass[n-1]=y1mass[0]=y2mass[0]=y3mass[0]=-1
    
    # return b1mass,b2mass,b3mass,y1mass,y2mass,y3mass
    return b1mass,b2mass,y1mass,y2mass


# seq='DSGKELHINLIPNKQ'
# mods=[]
# ions=calfragmass4regular(seq,mods)
# for i in ions:
#     print(i)


# 给定1+的mh，计算任意价态的m/z
def cal_mz(mh,charge):
    if mh<0:
        return -1
    else:
        return (mh+PMass*(charge-1))/charge

# 计算可碎裂交联肽段的单条肽段的碎片离子质量
# 交联剂断裂之后，相当于计算单肽的碎片离子质量
def calfragmass4clv(seq,LinkerName,linksite,mods={}):
    long_mass, short_mass=mpClvLinkerLongShortMass[LinkerName]
    n=len(seq)
    mass=[0.0]*n
    for i,aa in enumerate(seq):
        mass[i]=n14aamass[aa]
    for msite,mname in mods.items():
        mass[msite]+=(mpModMassN14[mname])
    
    #----b ion----
    linear_b1mass=[0.0]*n
    linear_b1mass[0]=mass[0]+PMass
    for i in range(1,n):
        linear_b1mass[i]=linear_b1mass[i-1]+mass[i]

    clv_long_b1mass=[0.0]*n
    clv_short_b1mass=[0.0]*n
    for i in range(n):
        if i>=linksite:
            clv_long_b1mass[i]=linear_b1mass[i]+long_mass
            clv_short_b1mass[i]=linear_b1mass[i]+short_mass
            linear_b1mass[i]=-1
        else:
            clv_long_b1mass[i]=-1
            clv_short_b1mass[i]=-1

    #---y ion---
    linear_y1mass=[0.0]*n
    linear_y1mass[n-1]=mass[n-1]+H2OMass+PMass
    for i in range(n-2,-1,-1):
        linear_y1mass[i]=linear_y1mass[i+1]+mass[i]
    

    clv_long_y1mass=[0.0]*n
    clv_short_y1mass=[0.0]*n
    for i in range(n):
        if i <= linksite:
            clv_long_y1mass[i]=linear_y1mass[i]+long_mass
            clv_short_y1mass[i]=linear_y1mass[i]+short_mass
            linear_y1mass[i]=-1
        else:
            clv_long_y1mass[i]=-1
            clv_short_y1mass[i]=-1
    # 2+
    linear_b2mass=[0.0]*n
    clv_long_b2mass=[0.0]*n
    clv_short_b2mass=[0.0]*n

    linear_y2mass=[0.0]*n
    clv_long_y2mass=[0.0]*n
    clv_short_y2mass=[0.0]*n

    for i in range(n):
        linear_b2mass[i]=cal_mz(linear_b1mass[i],2)
        linear_y2mass[i]=cal_mz(linear_y1mass[i],2)

        clv_long_b2mass[i]=cal_mz(clv_long_b1mass[i],2)
        clv_long_y2mass[i]=cal_mz(clv_long_y1mass[i],2)

        clv_short_b2mass[i]=cal_mz(clv_short_b1mass[i],2)
        clv_short_y2mass[i]=cal_mz(clv_short_y1mass[i],2)

    linear_b1mass[n-1]=linear_b2mass[n-1]=linear_y1mass[0]=linear_y2mass[0]=-1
    clv_long_b1mass[n-1]=clv_long_b2mass[n-1]=clv_long_y1mass[0]=clv_long_y2mass[0]=-1
    clv_short_b1mass[n-1]=clv_short_b2mass[n-1]=clv_short_y1mass[0]=clv_short_y2mass[0]=-1
    
    return linear_b1mass,linear_b2mass,linear_y1mass,linear_y2mass,clv_long_b1mass,clv_long_b2mass,clv_long_y1mass,clv_long_y2mass,clv_short_b1mass,clv_short_b2mass,clv_short_y1mass,clv_short_y2mass


# seq='GLSDGEWQQVLNVWGK'
# ions=calfragmass4clv(seq,'DSSO',15)
# for i in ions:
#     print(i)



def calonepep4xl(seq,mods,linksite,anothermass,LinkerName):
    LinkerXLMass=mpLinkerXLMass[LinkerName]
    n=len(seq)
    mass=[0.0]*n
    for i,aa in enumerate(seq):
        mass[i]=n14aamass[aa]
        
    try:
        for msite,mname in mods.items():
            mass[msite]+=(mpModMassN14[mname])
        mass[linksite]+=(LinkerXLMass+anothermass)
    except IndexError: #可能交联位点错误或者修饰位点错误
        print('linksite error, seq=%s,linksite=%d,mods=%s'%(seq,linksite,str(mods)))
    
    b1mass=[0.0]*n
    b1mass[0]=mass[0]+PMass
    for i in range(1,n):
        b1mass[i]=b1mass[i-1]+mass[i]
    y1mass=[0.0]*n
    y1mass[n-1]=mass[n-1]+H2OMass+PMass
    for i in range(n-2,-1,-1):
        y1mass[i]=y1mass[i+1]+mass[i]
        
    b2mass=[0.0]*n
    b3mass=[0.0]*n
    y2mass=[0.0]*n
    y3mass=[0.0]*n
    for i in range(n):
        b2mass[i]=(b1mass[i]+PMass)/2
        b3mass[i]=(b1mass[i]+2*PMass)/3
        
        y2mass[i]=(y1mass[i]+PMass)/2
        y3mass[i]=(y1mass[i]+2*PMass)/3
        
        
    b1mass[n-1]=b2mass[n-1]=b3mass[n-1]=y1mass[0]=y2mass[0]=y3mass[0]=-1
    
    return b1mass,b2mass,b3mass,y1mass,y2mass,y3mass


def calfragmass4xl(seq1,mods1,linksite1,seq2,mods2,linksite2,LinkerName):
    n14mass1=calpepmass(seq1)
    for msite,mname in mods1.items():
        n14mass1+=mpModMassN14[mname]
        
    n14mass2=calpepmass(seq2)
    for msite,mname in mods2.items():
        n14mass2+=mpModMassN14[mname]
    
    b1mass1,b2mass1,b3mass1,y1mass1,y2mass1,y3mass1=calonepep4xl(seq1,mods1,linksite1,n14mass2,LinkerName)
    b1mass2,b2mass2,b3mass2,y1mass2,y2mass2,y3mass2=calonepep4xl(seq2,mods2,linksite2,n14mass1,LinkerName)
    
    # return b1mass1,b2mass1,b3mass1,y1mass1,y2mass1,y3mass1,b1mass2,b2mass2,b3mass2,y1mass2,y2mass2,y3mass2
    return b1mass1,b2mass1,y1mass1,y2mass1,b1mass2,b2mass2,y1mass2,y2mass2



# seq1='GKSDNVPSEEVVK'
# mods1={}
# linksite1=1
# seq2='VKYVTEGMR'
# mods2={7:'Oxidation[M]'}
# linksite2=1
# LinkerName='DSS'

# ions=calfragmass4xl(seq1,mods1,linksite1,seq2,mods2,linksite2,LinkerName)
# for i in ions:
#     print(i)



# from .model import BiLSTMLinearPredictor # 相对路径导入
from pDeepXL.model import BiLSTMLinearPredictor # 绝对路径导入，两者都可以。
import pDeepXL.utils as utils

import configparser
import pickle
from tqdm import tqdm
import numpy as np
import time

import torch
import torch.autograd as autograd
import torch.nn as nn
import torch.optim as optim
import torch.utils.data
from torch.utils.data import DataLoader

import matplotlib
matplotlib.use('AGG')#或者PDF, SVG或PS
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.ticker import MaxNLocator
from sklearn.model_selection import train_test_split
import torch.nn.utils.rnn as rnn_utils


device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

config = configparser.ConfigParser()
path_config=r'pDeepXL/configs/config.ini'
# if os.path.isfile(path_config):
#     print('config file exist')
# else:
#     print('config file not exist')
config.read(path_config, encoding='UTF-8')

MIN_PREC_CHARGE=int(config['DEFAULT']['min_prec_charge'])
MAX_PREC_CHARGE=int(config['DEFAULT']['max_prec_charge'])
MIN_PEPTIDE_LEN=int(config['DEFAULT']['min_peptide_len'])
MAX_PEPTIDE_LEN=int(config['DEFAULT']['max_peptide_len'])

from pDeepXL.featurer import PeptideFeaturer
pf = PeptideFeaturer(config)


def GetPeptideFeatures(CSMs, consider_xlink_ion):
    T,L,X=[],[],[] # L表示总长度
    PT=[] # 肽段类型peptide type，0 for linear, 1 for non-clv, 2 for clv
    L1,L2,S1,S2=[],[],[],[] # α和β的长度，及交联位点，如果是单肽（P=0），则这几个值都是-1（无效）

    cur_pt=1 if consider_xlink_ion else 2

    num_invalid=0
    for csm in tqdm(CSMs):
        title,scan,prec_charge,instrument,NCE_low, NCE_medium, NCE_high,LinkerName,seq1,mods1,linksite1,seq2,mods2,linksite2=csm
        l1,l2 = len(seq1),len(seq2)

        if not (prec_charge >= MIN_PREC_CHARGE and prec_charge <= MAX_PREC_CHARGE \
                and l1 >= MIN_PEPTIDE_LEN and l1 <= MAX_PEPTIDE_LEN \
                and l2 >= MIN_PEPTIDE_LEN and l2 <= MAX_PEPTIDE_LEN): # 电荷与长度的约束
            num_invalid+=1
            continue

        seq1_vec,seq1_aa_mod_sum = pf.Sequence2Vec(seq1,mods1,prec_charge,instrument,NCE_low, NCE_medium, NCE_high,LinkerName,linksite1,False)
        seq2_vec,seq2_aa_mod_sum = pf.Sequence2Vec(seq2,mods2,prec_charge,instrument,NCE_low, NCE_medium, NCE_high,LinkerName,linksite2,False)

        # 对于不可碎裂交联剂，累加和需要考虑互补肽段
        if consider_xlink_ion:
            seq1_vec = pf.AddAnotherSeq(seq1_vec, linksite1, seq2_aa_mod_sum)
            seq2_vec = pf.AddAnotherSeq(seq2_vec, linksite2, seq1_aa_mod_sum)

        linker_symbol_vec = [pf.LinkerSymbol2Vec()]

        X.append(seq1_vec+linker_symbol_vec+seq2_vec)
        L.append(l1-1+l2-1+1)
        T.append(title)

        PT.append(cur_pt)
        L1.append(l1-1)
        L2.append(l2-1)
        S1.append(linksite1)
        S2.append(linksite2)

    print('collected %d valid samples, discard %d invalid samples'%(len(T),num_invalid))
    return T,PT,L,L1,L2,S1,S2,X


def LoadPredictDataSet(BATCH_SIZE, consider_xlink_ion, path_data_file):
    CSMs,mpTitleLines,header = utils.ReadpLink2PredictFile(path_data_file)
    T,PT,L,L1,L2,S1,S2,X=GetPeptideFeatures(CSMs, consider_xlink_ion)
    predict_data_loader=utils.ConvertPredictDataToDataLoader(BATCH_SIZE, X, T, PT, L1, L2, S1, S2)
    return predict_data_loader,mpTitleLines,header


def PredictMS2(model, iterator):
    model.eval()
    mpPeaks={}
    mpSpecInfo={}

    for batch_x, batch_length,batch_title,batch_pt,batch_l1,batch_l2,batch_s1,batch_s2 in iterator:

        batch_x_pack = rnn_utils.pack_padded_sequence(batch_x, batch_length, batch_first=True)
        batch_x_pack = batch_x_pack.to(device)
        masked_y_pred,out_len = model.work(batch_x_pack)

        for title, pred, lseq, pt,l1,l2,s1,s2 in zip(batch_title,masked_y_pred,out_len, batch_pt,batch_l1,batch_l2,batch_s1,batch_s2):
            peaks=[]
            for i in range(lseq):
                peaks.append(pred[i].cpu().detach().numpy())
            peaks=np.array(peaks)
            peaks=np.transpose(peaks)
            mpPeaks[title]=peaks
            mpSpecInfo[title]=[pt,l1,l2,s1,s2]
    return mpPeaks, mpSpecInfo



def get_model(is_non_cleavable):
    INPUT_DIM = 130
    HIDDEN_DIM = 256
    OUTPUT_DIM = 8
    N_LAYERS = 2
    BIDIRECTIONAL = True
    DROPOUT = 0.5
    path_model=r'pDeepXL/pt/non_clv_model.pt' if is_non_cleavable else r'pDeepXL/pt/clv_model.pt'

    print('loading pretrained model...')
    model = BiLSTMLinearPredictor(INPUT_DIM, HIDDEN_DIM, OUTPUT_DIM, N_LAYERS, BIDIRECTIONAL, DROPOUT)
    model.load_state_dict(torch.load(path_model,map_location=device))
    model = model.to(device)

    return model


# ----- single predict -----------

def predict_single(prec_charge,instrument,NCE_low, NCE_medium, NCE_high,crosslinker,seq1,mods1,linksite1,seq2,mods2,linksite2):
    
    if crosslinker=='DSSO' or crosslinker=='DSBU':
        is_non_cleavable=False
    elif crosslinker=='DSS' or crosslinker=='Leiker':
        is_non_cleavable=True
    else:
        print('sorry, pDeepXL does not support %s cross-linker'%crosslinker)
        return

    #--- preprocess data ---
    title='anonymous.dta'
    scan='0'
    single_csm=[title,scan,prec_charge,instrument,NCE_low, NCE_medium, NCE_high,crosslinker,seq1,mods1,linksite1,seq2,mods2,linksite2]
    CSMs=[single_csm]

    T,PT,L,L1,L2,S1,S2,X=GetPeptideFeatures(CSMs, is_non_cleavable)
    predict_data_loader=utils.ConvertPredictDataToDataLoader(1, X, T, PT, L1, L2, S1, S2) # batchsize=1
    #----------------------

    model = get_model(is_non_cleavable)

    print('predicting...')
    mpPeaks, mpSpecInfo=PredictMS2(model, predict_data_loader)
    print('prediction done')
    
    # 把预测输出格式转换为plot的输入格式
    peaks = mpPeaks[title]
    pt,l1,l2,s1,s2=mpSpecInfo[title] # 这里记录的l1和l2是序列的真实长度-1
    alpha_peaks=peaks[:,:l1].tolist()
    beta_peaks=peaks[:,l1+1:].tolist() # 跳过中间的交联剂输出

    # 与dta的格式一致，补齐无效的b离子和y离子，使得离子长度和肽段长度一致
    for idx in range(8):
        if idx%4<=1: # b ion
            alpha_peaks[idx].append(0.0) 
            beta_peaks[idx].append(0.0)
        else: # y ion
            alpha_peaks[idx].insert(0,0.0)
            beta_peaks[idx].insert(0,0.0)

    pred_matrix=alpha_peaks,beta_peaks # plot_single接收的输入格式

    return mpPeaks, mpSpecInfo,pred_matrix

#----------------------------------------------------------------




# ----- batch predict and save to file -----------

def predict_batch(path_data_file,is_non_cleavable):

    BATCH_SIZE=1024
    print('loading predicting data...')
    predict_data_loader,mpTitleLines,header=LoadPredictDataSet(BATCH_SIZE, is_non_cleavable, path_data_file)

    model = get_model(is_non_cleavable)

    print('predicting...')
    mpPeaks, mpSpecInfo=PredictMS2(model, predict_data_loader)
    print('prediction done')
    
    return mpPeaks, mpTitleLines, header, mpSpecInfo


def save_result_batch(path, predicted_results):
    mpPeaks, mpTitleLines, header, mpSpecInfo=predicted_results
    fout=open(path,'w')
    fout.write(header+'\t')
    fout.write('seq1_pred_b1b2y1y2\tseq2_pred_b1b2y1y2\n')
    for title, peaks in mpPeaks.items():
        fout.write(mpTitleLines[title]+'\t')
        pt,l1,l2,s1,s2=mpSpecInfo[title] # 这里记录的l1和l2是序列的真实长度-1
        alpha_peaks=peaks[:,:l1].tolist()
        beta_peaks=peaks[:,l1+1:].tolist() # 跳过中间的交联剂输出

        # 与dta的格式一致，补齐无效的b离子和y离子，使得离子长度和肽段长度一致
        for idx in range(8):
            if idx%4<=1: # b ion
                alpha_peaks[idx].append(0.0) 
                beta_peaks[idx].append(0.0)
            else: # y ion
                alpha_peaks[idx].insert(0,0.0)
                beta_peaks[idx].insert(0,0.0)

        fout.write('\t'.join(map(str,[alpha_peaks]))+'\t')
        fout.write('\t'.join(map(str,[beta_peaks]))+'\n')
    fout.close()
    print('write done')

#----------------------------------------------------------------



import torch
import torch.autograd as autograd
import torch.nn as nn
import torch.optim as optim
import torch.nn.utils.rnn as rnn_utils
import torch.nn.functional as F

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

class BiLSTMLinearPredictor(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim, n_layers, bidirectional, dropout):
        super().__init__()
        
        self.lstm = nn.LSTM(input_dim, hidden_dim, num_layers = n_layers, bidirectional = bidirectional, dropout = dropout, batch_first = True)
        self.fc1 = nn.Linear(hidden_dim * 2 if bidirectional else hidden_dim, output_dim)
        self.fc2 = nn.ReLU(inplace=True)

    # https://blog.csdn.net/Mr_green_bean/article/details/104713382
    # 肽段类型peptide type，0 for linear, 1 for non-clv, 2 for clv
    def mask_mae_loss(self, padded_y_truth, y_pred, y_length, pep_types):

        if pep_types[0]!=2: # linear or non-clv，只取前4维计算loss、更新梯度，不影响clv的transfer，非常重要！
            padded_y_truth=padded_y_truth[:,:,:4]
            y_pred=y_pred[:,:,:4]

        mask = torch.zeros(padded_y_truth.shape)
        for i,l in enumerate(y_length):
            mask[i,:l,:]=1
        mask = mask.to(device)
        masked_y_pred = y_pred.mul(mask)
        return F.l1_loss(masked_y_pred, padded_y_truth)

    def work(self, packed_peptides):
        outputs, (hidden, cell) = self.lstm(packed_peptides)
        out_unpack, out_len = rnn_utils.pad_packed_sequence(outputs, batch_first=True)
        predictions = self.fc2(self.fc1(out_unpack))
        
        mask = torch.zeros(predictions.shape)
        for i,l in enumerate(out_len):
            mask[i,:l,:]=1
        mask = mask.to(device)
        masked_y_pred = predictions.mul(mask)

        return masked_y_pred,out_len
    
    def forward(self, packed_peptides, padded_y_truth, pep_types):

        # padded_peptides shape: [batch size, peptide len, input dim]
        outputs, (hidden, cell) = self.lstm(packed_peptides)

        out_unpack, out_len = rnn_utils.pad_packed_sequence(outputs, batch_first=True)
        
        # predictions shape: [batch size, peptide len,  output dim]
        predictions = self.fc2(self.fc1(out_unpack))
        loss = self.mask_mae_loss(padded_y_truth,predictions,out_len,pep_types)
        
        return predictions,loss